/*
 * RT-FIFO devices
 *
 * 1997, Michael Barabanov
 *
 * A lot of this code is stolen from fs/pipe.c
 *
 */

#include <linux/module.h>
#include <linux/major.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/malloc.h>
#include <rtl_fifo.h>

#include <asm/system.h>
#include <rtl_sync.h>

#define RTF_NO 64

struct rt_fifo_struct {
	int open;
	int user_open;
	int sleeps;
	char *base;
	int bufsize;
	int start;
	int len;
	int (*user_handler) (unsigned int fifo);
	struct wait_queue *wait;
	struct tq_struct wake_up_task;
};

static struct rt_fifo_struct fifoes[RTF_NO];

#define RTF_ADDR(minor)		(&fifoes[minor])

#define RTF_OPEN(minor)		(RTF_ADDR(minor)->open)
#define RTF_USER_OPEN(minor)	(RTF_ADDR(minor)->user_open)
#define RTF_BASE(minor)		(RTF_ADDR(minor)->base)
#define RTF_BUF(minor)		(RTF_ADDR(minor)->bufsize)
#define RTF_START(minor)	(RTF_ADDR(minor)->start)
#define RTF_HANDLER(minor)	(RTF_ADDR(minor)->user_handler)
#define RTF_LEN(minor)		(RTF_ADDR(minor)->len)
#define RTF_FREE(minor)		(RTF_BUF(minor) - RTF_LEN(minor))
#define RTF_WAIT(minor)		(RTF_ADDR(minor)->wait)
#define RTF_SLEEPS(minor)	(RTF_ADDR(minor)->sleeps)
#define RTF_WAKE_UP_TASK(minor)	(RTF_ADDR(minor)->wake_up_task)

#define RTF_WRAP(minor,pos)	((pos) < RTF_BUF(minor)? (pos) : (pos) - RTF_BUF(minor))
#define RTF_END(minor)		RTF_WRAP(minor, RTF_START(minor)+RTF_LEN(minor))
#define RTF_EMPTY(minor)	(RTF_LEN(minor)==0)
#define RTF_FULL(minor)		(RTF_FREE(minor)==0)
#define RTF_MAX_RCHUNK(minor)	(RTF_BUF(minor) - RTF_START(minor))
#define RTF_MAX_WCHUNK(minor)	(RTF_BUF(minor) - RTF_END(minor))


static void rtf_wake_up(void *p)
{
	struct rt_fifo_struct *fifo_ptr = (struct rt_fifo_struct *) p;
	wake_up_interruptible(&(fifo_ptr->wait));
}


/* These are for use in real-time processes */

int rtf_resize(unsigned int minor, int size)
{
	void *mem;
	void *old;
	long interrupt_state;

	if (minor >= RTF_NO) {
		return -ENODEV;
	}
	mem = kmalloc(size, GFP_KERNEL);
	if (!mem) {
		return -ENOMEM;
	}
	old = RTF_BASE(minor);

	rtl_no_interrupts(interrupt_state);

	RTF_BASE(minor) = mem;
	RTF_BUF(minor) = size;
	RTF_START(minor) = 0;
	RTF_LEN(minor) = 0;

	rtl_restore_interrupts(interrupt_state);

	if (RTF_OPEN(minor)) {
		kfree(old);
	}
	return size;
}


int rtf_create(unsigned int minor, int size)
{
	int result;

	if (minor >= RTF_NO) {
		return -ENODEV;
	}
	if (RTF_OPEN(minor)) {
		return -EBUSY;
	}
	if ((result = rtf_resize(minor, size)) < 0) {
		return result;
	}
	RTF_OPEN(minor) = 1;
	RTF_SLEEPS(minor) = 0;
	RTF_USER_OPEN(minor) = 0;
	RTF_HANDLER(minor) = 0;

	RTF_WAKE_UP_TASK(minor).next = 0;
	RTF_WAKE_UP_TASK(minor).sync = 0;
	RTF_WAKE_UP_TASK(minor).data = RTF_ADDR(minor);
	RTF_WAKE_UP_TASK(minor).routine = &rtf_wake_up;

	RTF_WAIT(minor) = 0;

	return 0;
}


int rtf_destroy(unsigned int minor)
{
	if (minor >= RTF_NO) {
		return -ENODEV;
	}
	if (RTF_USER_OPEN(minor)) {
		return -EINVAL;
	}
	if (!RTF_OPEN(minor)) {
		return -EINVAL;
	}
	RTF_OPEN(minor) = 0;
	kfree(RTF_BASE(minor));
	return 0;
}


int rtf_create_handler(unsigned int minor, int (*handler) (unsigned int fifo))
{
	if (minor >= RTF_NO || !RTF_OPEN(minor) || !handler) {
		return -EINVAL;
	}
	RTF_HANDLER(minor) = handler;
	return 0;
}


int rtf_put(unsigned int minor, void *buf, int count)
{
	long interrupt_state;
	int chars = 0, free = 0, written = 0;
	char *pipebuf;

	if (minor >= RTF_NO) {
		return -ENODEV;
	}
	if (!RTF_OPEN(minor))
		return -EINVAL;

	rtl_no_interrupts(interrupt_state);
	if (RTF_FREE(minor) < count) {
		rtl_restore_interrupts(interrupt_state);
		return -ENOSPC;
	}
	while (count > 0 && (free = RTF_FREE(minor))) {
		chars = RTF_MAX_WCHUNK(minor);
		if (chars > count)
			chars = count;
		if (chars > free)
			chars = free;
		pipebuf = RTF_BASE(minor) + RTF_END(minor);
		written += chars;
		RTF_LEN(minor) += chars;
		count -= chars;
		memcpy(pipebuf, buf, chars);
		buf += chars;
	}
	rtl_restore_interrupts(interrupt_state);
	if (clear_bit(1, &RTF_SLEEPS(minor))) {
		queue_task_rt(&RTF_WAKE_UP_TASK(minor), &tq_rt);
	}
	return count;
}


int rtf_get(unsigned int minor, void *buf, int count)
{
	long interrupt_state;
	int chars = 0, size = 0, read = 0;
	char *pipebuf;

	if (minor >= RTF_NO) {
		return -ENODEV;
	}
	if (!RTF_OPEN(minor))
		return -EINVAL;

	rtl_no_interrupts(interrupt_state);
	while (count > 0 && (size = RTF_LEN(minor))) {
		chars = RTF_MAX_RCHUNK(minor);
		if (chars > count)
			chars = count;
		if (chars > size)
			chars = size;

		read += chars;
		pipebuf = RTF_BASE(minor) + RTF_START(minor);
		RTF_START(minor) += chars;
		RTF_START(minor) = RTF_WRAP(minor, RTF_START(minor));
		RTF_LEN(minor) -= chars;
		count -= chars;
		memcpy(buf, pipebuf, chars);
		buf += chars;
	}
	rtl_restore_interrupts(interrupt_state);
	if (clear_bit(1, &RTF_SLEEPS(minor))) {
		queue_task_rt(&RTF_WAKE_UP_TASK(minor), &tq_rt);
	}
	return read;
}


/* 
 * these are file_operations functions
 */

static int rtf_open(struct inode *inode, struct file *filp)
{
	unsigned int minor = MINOR(inode->i_rdev);

	if (minor >= RTF_NO)
		return -ENODEV;

	if (!RTF_OPEN(minor)) {
		return -EINVAL;
	}
	if (RTF_USER_OPEN(minor)) {
		return -EBUSY;
	}
	RTF_USER_OPEN(minor) = 1;

	MOD_INC_USE_COUNT;
	return 0;
}


static void rtf_release(struct inode *inode, struct file *file)
{
	unsigned int minor = MINOR(inode->i_rdev);

	if (!RTF_USER_OPEN(minor)) {
		printk("rtf: release on a not opened descriptor\n");
		return;
	}
	RTF_SLEEPS(minor) = 0;
	RTF_USER_OPEN(minor) = 0;
	MOD_DEC_USE_COUNT;
}


static int rtf_lseek(struct inode *inode, struct file *file, off_t offset, int origin)
{
	return -ESPIPE;
}


static int rtf_read(struct inode *inode, struct file *filp, char *buf, int count)
{
	unsigned int minor = MINOR(inode->i_rdev);
	long interrupt_state;
	int result;

	int chars = 0, size = 0, read = 0;
	char *pipebuf;

	if (filp->f_flags & O_NONBLOCK) {
		/*      if (RTF_LOCK(minor))
		   return -EAGAIN;
		 */
		if (RTF_EMPTY(minor))
			return 0;
	} else
		while (RTF_EMPTY(minor) /* || RTF_LOCK(minor) */ ) {
			if (current->signal & ~current->blocked)
				return -ERESTARTSYS;
			set_bit(1, &RTF_SLEEPS(minor));
			interruptible_sleep_on(&RTF_WAIT(minor));
		}
/*      RTF_LOCK(minor)++; */
	while (count > 0 && (size = RTF_LEN(minor))) {
		chars = RTF_MAX_RCHUNK(minor);
		if (chars > count)
			chars = count;
		if (chars > size)
			chars = size;

		read += chars;
		pipebuf = RTF_BASE(minor) + RTF_START(minor);
		count -= chars;
		memcpy_tofs(buf, pipebuf, chars);
		rtl_no_interrupts(interrupt_state);
		RTF_START(minor) += chars;
		RTF_START(minor) = RTF_WRAP(minor, RTF_START(minor));
		RTF_LEN(minor) -= chars;
		rtl_restore_interrupts(interrupt_state);
		buf += chars;
	}
/*      RTF_LOCK(minor)--; */
	if (read) {
		inode->i_atime = CURRENT_TIME;
		if (RTF_HANDLER(minor)) {
			if ((result = (*RTF_HANDLER(minor)) (minor)) < 0) {
				return result;
			}
		}
		return read;
	}
	return 0;
}


static int rtf_write(struct inode *inode, struct file *filp, const char *buf, int count)
{
	unsigned int minor = MINOR(inode->i_rdev);
	long interrupt_state;
	int chars = 0, free = 0, written = 0;
	char *pipebuf;
	int result;

	if (count <= RTF_BUF(minor))
		free = count;
	else
		free = 1;

	while (count > 0) {
		while ((RTF_FREE(minor) < free) /* || RTF_LOCK(minor) */ ) {
			if (current->signal & ~current->blocked)
				return written ? : -ERESTARTSYS;
			if (filp->f_flags & O_NONBLOCK)
				return written ? : -EAGAIN;
			set_bit(1, &RTF_SLEEPS(minor));
			interruptible_sleep_on(&RTF_WAIT(minor));
		}
		/*      RTF_LOCK(minor)++; */
		while (count > 0 && (free = RTF_FREE(minor))) {
			chars = RTF_MAX_WCHUNK(minor);
			if (chars > count)
				chars = count;
			if (chars > free)
				chars = free;
			rtl_no_interrupts(interrupt_state);
			pipebuf = RTF_BASE(minor) + RTF_END(minor);
			rtl_restore_interrupts(interrupt_state);
			count -= chars;
			written += chars;

			memcpy_fromfs(pipebuf, buf, chars);

			rtl_no_interrupts(interrupt_state);
			RTF_LEN(minor) += chars;
			rtl_restore_interrupts(interrupt_state);
			buf += chars;
		}
		/*      RTF_LOCK(minor)--; */
		free = 1;
	}
	inode->i_ctime = inode->i_mtime = CURRENT_TIME;
	if (RTF_HANDLER(minor)) {
		if ((result = (*RTF_HANDLER(minor)) (minor)) < 0) {
			return result;
		}
	}
	return written;
}

static int rtf_select(struct inode *inode, struct file *filp, int sel_type, select_table * wait)
{
	unsigned int minor = MINOR(inode->i_rdev);
	switch (sel_type) {
		case SEL_IN:
			if (!RTF_EMPTY(minor))
				return 1;
			set_bit(1,&RTF_SLEEPS(minor));
			select_wait(&RTF_WAIT(minor), wait);
			return 0;
		case SEL_OUT:
			if (RTF_EMPTY(minor))
				return 1;
			set_bit(1, &RTF_SLEEPS(minor));
			select_wait(&RTF_WAIT(minor), wait);
			return 0;
	}
	return 0;
}


static struct file_operations rtf_fops =
{
	rtf_lseek,
	rtf_read,
	rtf_write,
	NULL,
	rtf_select,
	NULL,
	NULL,
	rtf_open,
	rtf_release
};



int rtf_init(void)
{
	int i;

	if (register_chrdev(RTF_MAJOR, "rtf", &rtf_fops)) {
		printk("RT-FIFO: unable to get major %d\n", RTF_MAJOR);
		return -EIO;
	}
	for (i = 0; i < RTF_NO; i++) {
		RTF_OPEN(i) = 0;
	}

	return 0;
}


#ifdef MODULE


int init_module(void)
{
	return rtf_init();
}


void cleanup_module(void)
{
	unregister_chrdev(RTF_MAJOR, "rtf");
}


#endif				/* MODULE */
