/*
 * RT-FIFO devices
 *
 * 1997, Michael Barabanov
 *
 * A lot of this code is stolen from fs/pipe.c
 *
 */
/* Named fifo code originally by Gavin Peters, gpeters@wireless-sys.com
   with changes by Steve Papacharalambous, stevepapa@mistral.co.uk ZentropixInc.
   All code copyright the authors, 1999, and released under the terms of the
   GNU LGPL.
   */

#include <linux/module.h>
#include <linux/major.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/malloc.h>
#include <linux/proc_fs.h>
#include <linux/string.h>
#include <linux/ctype.h>
#include <rtl_nfifo.h>

#include <asm/system.h>
#include <rtl_sync.h>

#define RTF_NO 64

/* 
 * Specify which RT-FIFO to start named creation.
 * RTFA_BEGIN .. RTF_NO
 */
#define RTFA_BEGIN 48
#define RTFA_NO ( RTF_NO - RTFA_BEGIN ) 

struct rt_fifo_struct {
	int open;
	int user_open;
	int sleeps;
	char *base;
	int bufsize;
	int start;
	int len;
	int (*user_handler) (unsigned int fifo);
	struct wait_queue *wait;
	struct tq_struct wake_up_task;
};


/* 
 * to keep track of allocations.  since we use an array in which the
 * fifo number is maintained implicitly, the fifo member is redundant,
 * but I put it there Just In Case(tm).
 */
struct rtfa_allocation {

	char* name;
  	unsigned int fifo;
};


static struct rt_fifo_struct fifoes[RTF_NO];


/*
 * allocated/unallocated named RT-FIFOS - NULL indicates unallocated
 */
static struct rtfa_allocation* allocs[ RTFA_NO ] = { NULL, };

/*
 * a doctored pointer to let us index by the fifo minor.
 */
static struct rtfa_allocation** allocs_ = allocs - RTFA_BEGIN;


/*
 * Debug flag.
 */
static int debug = 0;


/*
 * signatures of file (static) utility functions.
 */
static char is_unique_name(const char* name);
static int find_with_name(const char* name);
static int resolve_fifo_name(const char *base, char **name_op);


#define RTF_ADDR(minor)		(&fifoes[minor])

#define RTF_OPEN(minor)		(RTF_ADDR(minor)->open)
#define RTF_USER_OPEN(minor)	(RTF_ADDR(minor)->user_open)
#define RTF_BASE(minor)		(RTF_ADDR(minor)->base)
#define RTF_BUF(minor)		(RTF_ADDR(minor)->bufsize)
#define RTF_START(minor)	(RTF_ADDR(minor)->start)
#define RTF_HANDLER(minor)	(RTF_ADDR(minor)->user_handler)
#define RTF_LEN(minor)		(RTF_ADDR(minor)->len)
#define RTF_FREE(minor)		(RTF_BUF(minor) - RTF_LEN(minor))
#define RTF_WAIT(minor)		(RTF_ADDR(minor)->wait)
#define RTF_SLEEPS(minor)	(RTF_ADDR(minor)->sleeps)
#define RTF_WAKE_UP_TASK(minor)	(RTF_ADDR(minor)->wake_up_task)

#define RTF_WRAP(minor,pos)	((pos) < RTF_BUF(minor)? (pos) : (pos) - RTF_BUF(minor))
#define RTF_END(minor)		RTF_WRAP(minor, RTF_START(minor)+RTF_LEN(minor))
#define RTF_EMPTY(minor)	(RTF_LEN(minor)==0)
#define RTF_FULL(minor)		(RTF_FREE(minor)==0)
#define RTF_MAX_RCHUNK(minor)	(RTF_BUF(minor) - RTF_START(minor))
#define RTF_MAX_WCHUNK(minor)	(RTF_BUF(minor) - RTF_END(minor))


static void rtf_wake_up(void *p)
{
	struct rt_fifo_struct *fifo_ptr = (struct rt_fifo_struct *) p;
	wake_up_interruptible(&(fifo_ptr->wait));
}


/* These are for use in real-time processes */

int rtf_resize(unsigned int minor, int size)
{
	void *mem;
	void *old;
	long interrupt_state;

	if (minor >= RTF_NO) {
		return -ENODEV;
	}
	mem = kmalloc(size, GFP_KERNEL);
	if (!mem) {
		return -ENOMEM;
	}
	old = RTF_BASE(minor);

	rtl_no_interrupts(interrupt_state);

	RTF_BASE(minor) = mem;
	RTF_BUF(minor) = size;
	RTF_START(minor) = 0;
	RTF_LEN(minor) = 0;

	rtl_restore_interrupts(interrupt_state);

	if (RTF_OPEN(minor)) {
		kfree(old);
	}
	return size;
}


int rtf_create(unsigned int minor, int size)
{
	int result;

	if (minor >= RTF_NO) {
		return -ENODEV;
	}
	if (RTF_OPEN(minor)) {
		return -EBUSY;
	}
	if ((result = rtf_resize(minor, size)) < 0) {
		return result;
	}
	RTF_OPEN(minor) = 1;
	RTF_SLEEPS(minor) = 0;
	RTF_USER_OPEN(minor) = 0;
	RTF_HANDLER(minor) = 0;

	RTF_WAKE_UP_TASK(minor).next = 0;
	RTF_WAKE_UP_TASK(minor).sync = 0;
	RTF_WAKE_UP_TASK(minor).data = RTF_ADDR(minor);
	RTF_WAKE_UP_TASK(minor).routine = &rtf_wake_up;

	RTF_WAIT(minor) = 0;

	return 0;
}


int rtf_destroy(unsigned int minor)
{
	if (minor >= RTF_NO) {
		return -ENODEV;
	}
	if (RTF_USER_OPEN(minor)) {
		return -EINVAL;
	}
	if (!RTF_OPEN(minor)) {
		return -EINVAL;
	}
	RTF_OPEN(minor) = 0;
	kfree(RTF_BASE(minor));
	return 0;
}


int rtf_create_handler(unsigned int minor, int (*handler) (unsigned int fifo))
{
	if (minor >= RTF_NO || !RTF_OPEN(minor) || !handler) {
		return -EINVAL;
	}
	RTF_HANDLER(minor) = handler;
	return 0;
}


int rtf_put(unsigned int minor, void *buf, int count)
{
	long interrupt_state;
	int chars = 0, free = 0, written = 0;
	char *pipebuf;

	if (minor >= RTF_NO) {
		return -ENODEV;
	}
	if (!RTF_OPEN(minor))
		return -EINVAL;

	rtl_no_interrupts(interrupt_state);
	if (RTF_FREE(minor) < count) {
		rtl_restore_interrupts(interrupt_state);
		return -ENOSPC;
	}
	while (count > 0 && (free = RTF_FREE(minor))) {
		chars = RTF_MAX_WCHUNK(minor);
		if (chars > count)
			chars = count;
		if (chars > free)
			chars = free;
		pipebuf = RTF_BASE(minor) + RTF_END(minor);
		written += chars;
		RTF_LEN(minor) += chars;
		count -= chars;
		memcpy(pipebuf, buf, chars);
		buf += chars;
	}
	rtl_restore_interrupts(interrupt_state);
	if (clear_bit(1, &RTF_SLEEPS(minor))) {
		queue_task_rt(&RTF_WAKE_UP_TASK(minor), &tq_rt);
	}
	return count;
}


int rtf_get(unsigned int minor, void *buf, int count)
{
	long interrupt_state;
	int chars = 0, size = 0, read = 0;
	char *pipebuf;

	if (minor >= RTF_NO) {
		return -ENODEV;
	}
	if (!RTF_OPEN(minor))
		return -EINVAL;

	rtl_no_interrupts(interrupt_state);
	while (count > 0 && (size = RTF_LEN(minor))) {
		chars = RTF_MAX_RCHUNK(minor);
		if (chars > count)
			chars = count;
		if (chars > size)
			chars = size;

		read += chars;
		pipebuf = RTF_BASE(minor) + RTF_START(minor);
		RTF_START(minor) += chars;
		RTF_START(minor) = RTF_WRAP(minor, RTF_START(minor));
		RTF_LEN(minor) -= chars;
		count -= chars;
		memcpy(buf, pipebuf, chars);
		buf += chars;
	}
	rtl_restore_interrupts(interrupt_state);
	if (clear_bit(1, &RTF_SLEEPS(minor))) {
		queue_task_rt(&RTF_WAKE_UP_TASK(minor), &tq_rt);
	}
	return read;
}


int rtf_named_create( unsigned int *fifo,
	       const char *name, 
	       int size )
{
  	int openfifo;
	struct rtfa_allocation* alloc;
	int createresult;
	
	/* look for an unused fifo, and die if there isn't one */
	if( debug )
	        printk( "rtfa: Allocating a fifo to \"%s\"\n", name );
	openfifo = 0;
	while( allocs[ openfifo ] != NULL ) {
		openfifo++;

		if( openfifo >= RTFA_NO ) {
			return -ENODEV;
		}
	}

	/* 
	 * get and fill in a new structure
	 */

	alloc = (struct rtfa_allocation *) 
			kmalloc( sizeof( struct rtfa_allocation ), GFP_KERNEL );

	if( alloc == NULL ) {
		return -ENOMEM;
	}

	alloc->fifo = openfifo + RTFA_BEGIN;
        if( resolve_fifo_name( name, &alloc->name ) < 0 ) {
		if( debug )
			printk("rtfa: Cannot resolve name \"%s\"\n", name);
		createresult = -EEXIST;
		goto no_alloc_name;
	}  /* End if - Unable to resolve fifo name. */
	if( alloc->name == NULL ) {
		createresult = -ENOMEM;
		goto no_alloc_name;
	}  /* End if - Unable to allocate memory for fifo name. */
	if( debug )
		printk( "rtfa: It gets fifo %d and resolved name \"%s\"\n",
			alloc->fifo, alloc->name );

	/*
	 * make the fallthrough call, and either finalize or back off
	 * of our allocation depending on success/failure.
	 */

	createresult = rtf_create( alloc->fifo, size );
	if( debug )
		printk( "rtfa: rtf_create result %d\n", createresult );

	if( createresult == 0 ) {  /* success */

		if( debug )
			printk( "rtfa: success!\n" );
		allocs[ openfifo ] = alloc;
		*fifo = alloc->fifo;
		return 0;
	}

	if( debug )
		printk( "rtfa: failure!\n" );

	kfree( alloc->name );
 no_alloc_name:
	kfree( alloc );
	return createresult;
}

int rtf_named_destroy( unsigned int fifo )
{
  	int result;

	if( debug )
		printk( "rtfa: rtfa_destroy( %d)\n", fifo );

	if( allocs_[ fifo ] == NULL ) {
		if( debug )
			printk( "rtfa: we don't have it.  abort!\n" );
		return -ENODEV;
	}

	result = rtf_destroy( fifo );
	if( debug )
		printk( "rtfa: rtf_destroy( %d ) returns %d\n", fifo, result );
	if( result == 0 ) { /* success */
		kfree( allocs_[ fifo]->name );
	  	kfree( allocs_[ fifo ] );
	  	allocs_[ fifo ] = NULL;
		return 0;
	}
	else { /* failure */
	  	return result;
	}
}



/*
 * implementations of file (static) utility functions
 */


/* 
 * check the currently allocated fifo's to see if any have a given name,
 * and return zero if it's already taken, nonzero otherwise.
 */
static char is_unique_name(const char *name)
{
  return find_with_name( name ) == -1;
}

/* 
 * return which fifo uses a given name, or -1 if none do.
 */
static int find_with_name(const char *name)
{
	int i;

	for( i = 0; i < RTFA_NO; i++ ) {

	  	if( allocs[ i ] == NULL )
			continue;

		if( 0 != strcmp( allocs[ i ]->name, name ) )
		  	continue;
 
		/* we've continue'd if it's not a match */
		return i + RTFA_BEGIN;
	}

	/* we didn't find it */
	return -1;
}


/* -----------------------------------------------------------------------*/

/*
* Resolve the fifo name.  Check whether the requested fifo name is unique.
* If it is that's cool, allocate memory for the name, save it, and return
* success.  If not then append a decimal digit string to the base name
* and increment the string until we get a unique name.  (or run out of
* digits....)
*/
static int resolve_fifo_name(const char *base, char **name_op) {

	int name_found = 0;
	int namelen;
	int i;
	char *name;

	if( debug )
		printk("rtfa: resolve_fifo_name called with \"%s\"\n",base);

	/*
	 * allocate a new buffer - giving us 3 extra bytes, 2 for the
	 * decimal digits, and 1 for the null terminator.
	 */
	namelen = strlen( base );
	name = (char *)kmalloc( sizeof(char) * ( namelen + 3 ),
				GFP_KERNEL );
	if( name == NULL ) {
		*name_op = NULL;
		return 0;
	}

	strcpy( name, base );

	if( !is_unique_name( name ) ) {

		if( debug )
			printk("rtfa: resolve_fifo_name \"%s\" is not unique\n",name);
		/*
		 * The base is not unique.  Grr... Add a digit (up to a max
		 * of 2) and increment until we get a unique name, or
		 * if there's more then <base>99 return failure.
		 *
		 * Loop for up to 99 times to try and generate a unique
		 * name for the fifo.  This is limited to 99 as only 2
		 * characters can be added onto the end of the base string.
		 */
		for( i = 0; i <= 99; i++ ) {
			if( i > 9 ) {
				*(name + namelen) = i / 10 + '0';
				*(name + namelen + 1) = i % 10 + '0';
				*(name + namelen + 2) = '\0';
			}
			else {
				*(name + namelen) = i % 10 + '0';
				*(name + namelen + 1) = '\0';
			}
			if( is_unique_name( name ) ) {
				name_found = 1;
				break;
			}
		} /* End for loop - Try to generate a unique name */
		if( name_found == 0 ) {
			if ( debug )
				printk("rtfa: resolve_fifo_name failed to find unique name for \"%s\"\n", base);
			*name_op = NULL;
			kfree( name );
			return -1;
		} /* End if - Couldn't get a unique name. */

	} /* End if - base is not unique. */

	*name_op = name;
	if( debug )
		printk("rtfa: resolve_fifo_name name returned \"%s\"\n",*name_op);

	return 0;

} /* End function - resolve_fifo_name */


/* -----------------------------------------------------------------------*/


/* 
 * these are file_operations functions
 */

static int rtf_open(struct inode *inode, struct file *filp)
{
	unsigned int minor = MINOR(inode->i_rdev);

	if (minor >= RTF_NO)
		return -ENODEV;

	if (!RTF_OPEN(minor)) {
		return -EINVAL;
	}
	if (RTF_USER_OPEN(minor)) {
		return -EBUSY;
	}
	RTF_USER_OPEN(minor) = 1;

	MOD_INC_USE_COUNT;
	return 0;
}


static void rtf_release(struct inode *inode, struct file *file)
{
	unsigned int minor = MINOR(inode->i_rdev);

	if (!RTF_USER_OPEN(minor)) {
		printk("rtf: release on a not opened descriptor\n");
		return;
	}
	RTF_SLEEPS(minor) = 0;
	RTF_USER_OPEN(minor) = 0;
	MOD_DEC_USE_COUNT;
}


static int rtf_lseek(struct inode *inode, struct file *file, off_t offset, int origin)
{
	return -ESPIPE;
}


static int rtf_read(struct inode *inode, struct file *filp, char *buf, int count)
{
	unsigned int minor = MINOR(inode->i_rdev);
	long interrupt_state;
	int result;

	int chars = 0, size = 0, read = 0;
	char *pipebuf;

	if (filp->f_flags & O_NONBLOCK) {
		/*      if (RTF_LOCK(minor))
		   return -EAGAIN;
		 */
		if (RTF_EMPTY(minor))
			return 0;
	} else
		while (RTF_EMPTY(minor) /* || RTF_LOCK(minor) */ ) {
			if (current->signal & ~current->blocked)
				return -ERESTARTSYS;
			set_bit(1, &RTF_SLEEPS(minor));
			interruptible_sleep_on(&RTF_WAIT(minor));
		}
/*      RTF_LOCK(minor)++; */
	while (count > 0 && (size = RTF_LEN(minor))) {
		chars = RTF_MAX_RCHUNK(minor);
		if (chars > count)
			chars = count;
		if (chars > size)
			chars = size;

		read += chars;
		pipebuf = RTF_BASE(minor) + RTF_START(minor);
		count -= chars;
		memcpy_tofs(buf, pipebuf, chars);
		rtl_no_interrupts(interrupt_state);
		RTF_START(minor) += chars;
		RTF_START(minor) = RTF_WRAP(minor, RTF_START(minor));
		RTF_LEN(minor) -= chars;
		rtl_restore_interrupts(interrupt_state);
		buf += chars;
	}
/*      RTF_LOCK(minor)--; */
	if (read) {
		inode->i_atime = CURRENT_TIME;
		if (RTF_HANDLER(minor)) {
			if ((result = (*RTF_HANDLER(minor)) (minor)) < 0) {
				return result;
			}
		}
		return read;
	}
	return 0;
}


static int rtf_write(struct inode *inode, struct file *filp, const char *buf, int count)
{
	unsigned int minor = MINOR(inode->i_rdev);
	long interrupt_state;
	int chars = 0, free = 0, written = 0;
	char *pipebuf;
	int result;

	if (count <= RTF_BUF(minor))
		free = count;
	else
		free = 1;

	while (count > 0) {
		while ((RTF_FREE(minor) < free) /* || RTF_LOCK(minor) */ ) {
			if (current->signal & ~current->blocked)
				return written ? : -ERESTARTSYS;
			if (filp->f_flags & O_NONBLOCK)
				return written ? : -EAGAIN;
			set_bit(1, &RTF_SLEEPS(minor));
			interruptible_sleep_on(&RTF_WAIT(minor));
		}
		/*      RTF_LOCK(minor)++; */
		while (count > 0 && (free = RTF_FREE(minor))) {
			chars = RTF_MAX_WCHUNK(minor);
			if (chars > count)
				chars = count;
			if (chars > free)
				chars = free;
			rtl_no_interrupts(interrupt_state);
			pipebuf = RTF_BASE(minor) + RTF_END(minor);
			rtl_restore_interrupts(interrupt_state);
			count -= chars;
			written += chars;

			memcpy_fromfs(pipebuf, buf, chars);

			rtl_no_interrupts(interrupt_state);
			RTF_LEN(minor) += chars;
			rtl_restore_interrupts(interrupt_state);
			buf += chars;
		}
		/*      RTF_LOCK(minor)--; */
		free = 1;
	}
	inode->i_ctime = inode->i_mtime = CURRENT_TIME;
	if (RTF_HANDLER(minor)) {
		if ((result = (*RTF_HANDLER(minor)) (minor)) < 0) {
			return result;
		}
	}
	return written;
}

static int rtf_select(struct inode *inode, struct file *filp, int sel_type, select_table * wait)
{
	unsigned int minor = MINOR(inode->i_rdev);
	switch (sel_type) {
		case SEL_IN:
			if (!RTF_EMPTY(minor))
				return 1;
			set_bit(1,&RTF_SLEEPS(minor));
			select_wait(&RTF_WAIT(minor), wait);
			return 0;
		case SEL_OUT:
			if (RTF_EMPTY(minor))
				return 1;
			set_bit(1, &RTF_SLEEPS(minor));
			select_wait(&RTF_WAIT(minor), wait);
			return 0;
	}
	return 0;
}


static struct file_operations rtf_fops =
{
	rtf_lseek,
	rtf_read,
	rtf_write,
	NULL,
	rtf_select,
	NULL,
	NULL,
	rtf_open,
	rtf_release
};



int rtf_init(void)
{
	int i;

	if (register_chrdev(RTF_MAJOR, "rtf", &rtf_fops)) {
		printk("RT-FIFO: unable to get major %d\n", RTF_MAJOR);
		return -EIO;
	}
	for (i = 0; i < RTF_NO; i++) {
		RTF_OPEN(i) = 0;
	}

	return 0;
}


/* -----------------------------------------------------------------------*/


/*
 * for now, we're a single file in /proc, because pre 2.2 linux's suck.
 */
#define LIMIT	(PAGE_SIZE-80)  /* no printy printy after limit */

static int rtfa_read_procmem(char* buf, char** start, off_t offset,
			     int len, int unused)
{
	int i;
	
	len = 0;
	for( i=0; i < RTFA_NO; i++ ) {

	  	if( allocs[ i ] == NULL )  continue;

		len += sprintf( buf+len, "%d %s\n", 
				allocs[ i ]->fifo,
				allocs[ i ]->name );
		
		if( len > LIMIT )  return len;
	}
	for( i = 0; i < RTFA_BEGIN; i++ ) {

		if( !fifoes[i].open ) continue;

		len += sprintf( buf + len, "%d <noname>\n", i );

		if( len > LIMIT )  return len;
	}

	return len;
}

struct proc_dir_entry proc_rtf_entry = {
	0,
	3, "rtf",
	S_IFREG | S_IRUGO,
	1, 0, 0,
	0,
	NULL,
	&rtfa_read_procmem,
	/* nothing more */
};


/* -----------------------------------------------------------------------*/


#ifdef MODULE


int init_module(void)
{
  	int result;

  	result = proc_register_dynamic( &proc_root, &proc_rtf_entry );
	return rtf_init();
}


void cleanup_module(void)
{
	proc_unregister( &proc_root, proc_rtf_entry.low_ino );
	unregister_chrdev(RTF_MAJOR, "rtf");
}


#endif				/* MODULE */
