/* a primitive real-time scheduler 
copyright Victor Yodaiken, Michael Barabanov, VJY Associates LLC.
History: Barabanov
         modified Oleg Subbotin, Ramesh Nalluri, April 1998
         modified Victor Yodaiken, August 1998
             Added compatability macros
             changed search loop
	     replaced old style r_cli with rtl_no_interrupt
	     removed x86 code
10.23.98     modified by Michael Barabanov. Periodic mode works.
12.22.98     merged fp and non-fp schedulers again;
             implemented delayed FPU context switching
*/


/* This is the interrupt latency + scheduler latency,
   The user needs to determine this and hardcode it accordingly */


#define ADJUSTUP  10 /*if time to run is no more than this time ahead, run*/
#define ADJUSTDOWN 0 /*set timer to interrupt this time less than requested*/

#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/version.h>
#include <linux/errno.h>
#include <linux/malloc.h>
#include <linux/cons.h>

#include <asm/system.h>
#include <linux/timex.h>
#include <asm/rt_time.h>
#include <asm/rt_irq.h>
#include <rtl_sync.h>

static int linux_irq_state;
static int rtl_active = 0;
#define		rtl_make_unrtactive() { SFIF = linux_irq_state; rtl_active=0;}
#define 	rtl_make_rtactive() { if (!rtl_active) {linux_irq_state = SFIF; SFIF = 0; rtl_active=1;}}
#define 	rtl_is_rtactive() (rtl_active)


#include <linux/kernel.h>
#include <rtl_sched.h>


#ifdef RTL_DEBUG
void debug_resched(RT_TASK * ,RTIME );
#endif

#include <asm/segment.h>
#include "x86/switch.h"


static int rtl_periodic_mode;
static int rtl_period;
 
#define GETTIME() (rtl_periodic_mode ? rt_get_last_period_time() : rt_get_time())
 
RT_TASK *rtl_tasks;
static RT_TASK *rtl_zombies;

RT_TASK *rtl_current;
RT_TASK *rtl_task_fpu_owner;	/* the task whose FP context is currently in the FPU unit */

RT_TASK rtl_linux_task;

int rtl_startup(void (*fn)(int), int data)
{
	r_sti();
	(*fn)(data);
	rt_task_delete(rtl_current);
	
	/* should never reach this line */
	return 0;
}
	

int rt_task_init(RT_TASK *task, void (*fn)(int), int data, int stack_size, 
			int priority) 
{
	int *st;
	long interrupt_state;

	if (task->magic == RT_TASK_MAGIC) {
		return -EINVAL;
	}

	st = (int *) kmalloc(stack_size, GFP_KERNEL);	
	if (!st) {
		return -ENOMEM;
	}

	task->priority = priority; 
	task->magic = RT_TASK_MAGIC; 
	task->state = RT_TASK_DORMANT; 
	task->stack_bottom = st;
	task->stack = st + stack_size / sizeof(int); /* we divide by sizeof(int) because "st" is of type (int *) */
	task->uses_fp = 0;
	x86_init_stack(task,data,rtl_startup);
	rtl_no_interrupts(interrupt_state);
	task->next = rtl_tasks;
	rtl_tasks = task;
	rtl_restore_interrupts(interrupt_state);
	return 0;
}


/* this function should only be called in the Linux context */

static int rtl_delete_zombies(void)
{
	RT_TASK *task;
	long interrupt_state;

	rtl_no_interrupts(interrupt_state);
        
	while (rtl_zombies) {
/* 		conpr("z"); */
		task = rtl_zombies;
		rtl_zombies = task -> next;

		rtl_restore_interrupts (interrupt_state);

		kfree (task->stack_bottom);
		task -> magic = RT_TASK_MAGIC + 1; // for the next rt_task_init to succeed
		/* we can deallocate the task structure here if needed */

		rtl_no_interrupts (interrupt_state);
	}

	rtl_restore_interrupts (interrupt_state);
	return 0;
}


/* returns the "base" time of all periods */
RTIME rtl_set_periodic_mode(RTIME period)
{
	unsigned long interrupt_state;
	if (period <= RT_MAX_PERIOD) { /* for larger periods the periodic mode is not worth the trouble to implement */
		rtl_no_interrupts(interrupt_state);
		rtl_periodic_mode = 1;
		rtl_period = period;

		I8253_make_periodic(rtl_period);

		rtl_restore_interrupts(interrupt_state);
		return rt_get_last_period_time();
	}
	return RT_TIME_END;
}

int rtl_set_oneshot_mode(void)
{
	unsigned long interrupt_state;
	rtl_no_interrupts(interrupt_state);
	rtl_periodic_mode = 0;
	I8253_make_oneshot(0);
	rtl_restore_interrupts(interrupt_state);
	return 0;
}


void rtl_schedule(void)
{
	RTIME now;
	RT_TASK *t;
	RT_TASK *newt = 0; /* new task to run, if any */
	RT_TASK *nextt = 0; /* task that will preempt, if any */
	unsigned long interrupt_state;

	rtl_no_interrupts(interrupt_state);
	now = GETTIME();

        /* find a task that can be run  */
	for (t = rtl_tasks; t; t = t->next) {
		if( (t->state == RT_TASK_DELAYED)&&
				( t->resume_time < now + ADJUSTUP)){
		   t->state = RT_TASK_READY;
		}
		if(t->state == RT_TASK_READY)
			if (!newt || (t->priority < newt->priority)) {
				newt = t;
			}
	}
	/* find the pre-empting task */
	for (t = rtl_tasks; t; t = t->next) {
		if( (t->state == RT_TASK_DELAYED) &&
				(!newt || (t->priority < newt->priority)) &&
				(!nextt ||(t->resume_time < nextt->resume_time))
				)
			nextt = t;
	}

				

	/* DEBUG */
	{
		static int nolinux = 0;
		if(newt) nolinux++;
		else nolinux = 0;
		if(nolinux > 100000){
			conpr("HELP NO LINUX EVER!!!");
			nolinux = 0;
			newt->resume_time = now+100000;
		}
	}
	/* END DEBUG */


	/* if nextt there is a preemptor so set timer to preempt
	   if !nextt
	        if there is a rt task, turn off the timer
		if there is no rt task, linux needs a clock interrupt
		*/
	if (nextt) {
		rt_set_timer(nextt->resume_time - ADJUSTDOWN ); 
	} else {
	       	if (newt) { rt_no_timer(); }
		else { rt_set_timer(now+LATCH);}
	}

	/* if there is no rt task to run, then run linux */
	if (!newt){
		newt = &rtl_linux_task;
		rtl_make_unrtactive();
	} else {
		rtl_make_rtactive();
	}
	/* if the new task is the current task nothing to do */
	if (newt == rtl_current) {
		rtl_restore_interrupts(interrupt_state);
		return;
	} else	{ /*else switch out old, switch in new */

		rtl_switch_to(newt); /* switch the integer registers and the control flow */

		/* delay switching the FPU context until it is really needed */
		/* this trick works for uniprocessors and SMP with no RT-task migration (?) */
		if (rtl_current -> uses_fp && rtl_task_fpu_owner != rtl_current) {
			if (rtl_task_fpu_owner) {
/* 				conpr("s"); */
				rtl_fpu_save (rtl_task_fpu_owner);
			}
/* 			conpr("r"); */

			rtl_fpu_restore (rtl_current);
			rtl_task_fpu_owner = rtl_current;
		}
		rtl_restore_interrupts(interrupt_state);
	}
	if (rtl_is_linux_task (rtl_current)) {
		/* it is safe to do if we require that rt_task_delete() is called from Linux */
		/* but it's probably better to use task queues anyway */
		rtl_delete_zombies();
	}
}


int rt_task_use_fp (RT_TASK *task, int flag)
{
	task -> uses_fp = flag;
	if (flag) {
		rtl_task_init_fpu (task, rtl_task_fpu_owner);
	} else {
		if (rtl_task_fpu_owner == task) {
			rtl_task_fpu_owner = 0;
		}
	}

	return 0;
}


static void rtl_sched_timer_interrupt(void)
{
#ifdef RTL_DEBUG_TIME
	if (once) {
		conpr("timer_interrupt: optimistic:"); conprn(rt_get_last_period_time()); conpr("\n");
		conpr("timer_interrupt: real:"); conprn(rt_get_time()); conpr("\n");
		once = 0;
	}
#endif
	rtl_schedule();
}



/* should be called from a place where it is safe to use kfree()
 * just like rt_task_init();
 * 
 * if this restriction is a problem, we can use task queues */

int rt_task_delete(RT_TASK *task)
{
	RT_TASK *t;
	int found = 0;

	long interrupt_state;

	if (task->magic != RT_TASK_MAGIC) {
		return -EINVAL;
	}
	
	rtl_no_interrupts (interrupt_state);

	if (task != rtl_tasks) {
		for (t = rtl_tasks; t; t = t->next) {
			if (t->next == task) {
				t->next = task->next;
				found = 1;
				break;
			}
		}
		if (!found) {
			rtl_restore_interrupts(interrupt_state);
			return -EINVAL;
		}
	} else {
		rtl_tasks = task->next;
	}

	task->state = RT_TASK_ZOMBIE;
	task->next = rtl_zombies;
	rtl_zombies = task;

	if (rtl_task_fpu_owner == task) {
		rtl_task_fpu_owner = 0;
	}

	rtl_restore_interrupts (interrupt_state);

	if (rtl_is_linux_task (rtl_current)) {
		rtl_delete_zombies();
	}
	return 0;
}


int rt_task_suspend(RT_TASK *task)
{
	task->state = RT_TASK_DORMANT;
	rtl_schedule();
	return 0;
}


int rt_task_wakeup(RT_TASK *task)
{
	task->state = RT_TASK_READY;
	rtl_schedule();
	return 0;
}


int rt_task_wait(void)
{
	long interrupt_state;
	rtl_no_interrupts (interrupt_state);
	rtl_current->state = RT_TASK_DELAYED;
	rtl_current->resume_time += rtl_current->period; 
	rtl_schedule();
	rtl_restore_interrupts (interrupt_state);
	return 0;
}

int rt_task_make_periodic(RT_TASK *task, RTIME start_time, RTIME period)
{
	long interrupt_state;
	if (task->magic != RT_TASK_MAGIC) {
		return -EINVAL;
	}
	rtl_no_interrupts(interrupt_state);
	task->resume_time = start_time;
	task->period = period;
	task->state = RT_TASK_DELAYED;
	rt_set_timer (GETTIME());
	rtl_restore_interrupts(interrupt_state);
	return 0;
}

int init_module(void)
{
	long interrupt_state;
	rtl_no_interrupts(interrupt_state);
	rt_no_timer();
	rtl_tasks = 0;
	rtl_zombies = 0;
	rtl_current = &rtl_linux_task;
	rtl_linux_task . priority = RT_LOWEST_PRIORITY;
	rtl_linux_task . next = 0;
	rtl_linux_task . uses_fp = 1;
	rtl_linux_task . state = RT_TASK_READY;
	rtl_task_fpu_owner = &rtl_linux_task;

/* 	rtl_fpu_init(); */
	rtl_set_oneshot_mode ();
	rt_request_timer (&rtl_sched_timer_interrupt);
	rtl_restore_interrupts (interrupt_state);
	return 0;
}


void cleanup_module(void)
{
	RT_TASK *t;
	RT_TASK *next_t;
	rt_free_timer();

	/* actually if rtl_tasks is non-zero at this point,
	 * chances are the system has crashed already */
	for (t = rtl_tasks; t; t = next_t) {
		conpr("RTL: rtl_tasks != 0 in rtl_sched::cleanup_module !\n");
		next_t = t->next;
		if (t != &rtl_linux_task) {
			rt_task_delete(t);
		}
	}
	rtl_delete_zombies();
}


